/*Parser for GP

  Copyright (C) 2000-2006 Free Software Foundation, Inc.

  Author: Frank Heckenbach <frank@pascal.gnu.de>

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, version 2.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; see the file COPYING. If not, write to
  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  Boston, MA 02111-1307, USA. */

%option noyywrap
%option never-interactive
%option case-insensitive
%option nodefault
%pointer

%x Uses
%x AfterUses
%x UsesIn
%x Import1
%x Import
%x ImportIn
%x AfterImport
%x Export1
%x Export
%x AfterExport
%x Directive
%x Directive1
%x LinkerDirective
%x ResourceDirective

%{
/* Avoid some warnings with `-Wundef' and `-Wunused'. */
#define YY_MAIN 0
#define YY_STACK_USED 0
#define YY_ALWAYS_INTERACTIVE 0
/* Probably bug in flex 2.5.27. Don't know which version exactly. 2.5.4 does not
   have the problem and doesn't define YY_FLEX_SUBMINOR_VERSION. Probably a more
   precise version check would be better (but I don't want to try all versions). */
#if defined (YY_FLEX_SUBMINOR_VERSION) && YY_FLEX_SUBMINOR_VERSION < 31
static void *yy_start_stack = 0;
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cpas.h"

#define OPTIONS_ONLY
#define DEFINE_LANG_NAME(NAME)
#define GPC_OPT(DIRECTIVE, NAME, OPTION, VALUE, DESCRIPTION) NAME,
#define GPC_OPT0(DIRECTIVE, NAME, DESCRIPTION) NAME,
#define GPC_OPT1(DIRECTIVE, NAME, SETTING, DESCRIPTION) NAME,
const char *gpc_options[] = {
#include "lang-options.h"
  NULL };

static char *Buf, *SrcName, *ImportName, *ImportFile, **ModuleName, StringChar;
static int BufCount, BufPos, Column, Line, Comment, Digraph, ImplementationPos, PrevState, ComponentStarted, ModuleStatus, Parentheses;
static CStringList **ExportedInterfaces, **InterfaceUses, **ImplementationUses, **LinkedFiles, **ResourceFiles;
static TFileKind FileKind;

static void *xmalloc (size_t size)
{
  void *p = malloc (size);
  if (!p)
    Error ("out of memory");
  return p;
}

#define YY_INPUT(buf, result, max_size) (NextChar ((buf), &(result)))
static void NextChar (char *outbuf, int *result)
{
  char ch;
  int ok;
  do
    {
      if (BufPos == 0 || Buf[BufPos] == '\n')
        do
          {
            int Pos1 = BufPos;
            if (Pos1 != 0)
              Pos1++;
            if (Comment || StringChar != 0 || Buf[Pos1] != '#')
              break;
            BufPos = Pos1 + 1;
            while (Buf[BufPos] == ' ' || Buf[BufPos] == '\t')
              BufPos++;
            Pos1 = BufPos;
            while (Buf[BufPos] >= '0' && Buf[BufPos] <= '9')
              BufPos++;
            if (BufPos > Pos1)
              {
                Line = atoi (Buf + Pos1) - 1;
                while (BufPos < BufCount && Buf[BufPos] != '"' && Buf[BufPos] != '\n')
                  BufPos++;
                if (Buf[BufPos] == '"')
                  {
                    Pos1 = ++BufPos;
                    while (BufPos < BufCount && Buf[BufPos] != '"' && Buf[BufPos] != '\n')
                      BufPos++;
                    if (BufPos > Pos1)
                      {
                        free (SrcName);
                        SrcName = xmalloc (BufPos - Pos1 + 1);
                        strncpy (SrcName, Buf + Pos1, BufPos - Pos1);
                        SrcName[BufPos - Pos1] = 0;
                      }
                    while (BufPos < BufCount && Buf[BufPos] != '\n')
                      BufPos++;
                  }
              }
          }
        while (Buf[BufPos] == '\n');
      if (BufPos >= BufCount)
        {
          *result = YY_NULL;
          return;
        }
      ch = Buf[BufPos++];
      if (ch == '\n')
        {
          Column = 1;
          Line++;
        }
      else
        Column++;
      if (ch == '(' && Buf[BufPos] == '*')
        {
          BufPos++;
          Column++;
          ch = '{';
          Digraph = 1;
        }
      if (Digraph && ch == '*' && Buf[BufPos] == ')')
        {
          BufPos++;
          Column++;
          ch = '}';
          Digraph = 0;
        }
      ok = 1;
      if (StringChar != 0)
        {
          if (StringChar == '\\')
            StringChar = '"';
          else if (ch == StringChar)
            StringChar = 0;
          else if (StringChar == '"' && ch == '\\')
            StringChar = '\\';
        }
      else if (Comment)
        {
          if (ch == '}')
            Comment = 0;
          ok = 0;
        }
      else if (ch == '\'' || ch == '"')
        StringChar = ch;
      else if (ch == '{' && Buf[BufPos] != '$')
        {
          Comment = 1;
          ok = 0;
        }
    }
  while (!ok);
  outbuf[0] = ch;
  *result = 1;
}

static void SetImplementation ()
{
  ImplementationPos = BufPos;
}

static void ParseError (const char *) __attribute__ ((noreturn));
static void ParseError (const char *Msg)
{
  ErrorLineColumnCString (SrcName, Line, Column, Msg);
}

static void SetFileKind (TFileKind NewFileKind)
{
  if (ComponentStarted)
    return;
  if (FileKind != fk_None)
    ParseError ("multiple programs, units and/or modules in one file are not supported");
  FileKind = NewFileKind;
}

static void AddList (CStringList ***List, char *Name1, char *Name2)
{
  **List = xmalloc (sizeof (CStringList));
  (**List)->Next = NULL;
  (**List)->Name1 = strdup (Name1);
  (**List)->Name2 = Name2 ? strdup (Name2) : NULL;
  *List = &((**List)->Next);
}

/* @@ The trouble with `import'/`export' syntax. Fails if those dialect keywords are used as module names. */
static int starts_decl_part (const char *s)
{
  return !strcasecmp (s, "import") ||
         !strcasecmp (s, "uses") ||
         !strcasecmp (s, "const") ||
         !strcasecmp (s, "type") ||
         !strcasecmp (s, "var") ||
         !strcasecmp (s, "label") ||
         !strcasecmp (s, "procedure") ||
         !strcasecmp (s, "function") ||
         !strcasecmp (s, "operator") ||
         !strcasecmp (s, "constructor") ||
         !strcasecmp (s, "destructor") ||
         !strcasecmp (s, "implementation") ||
         !strcasecmp (s, "begin") ||
         !strcasecmp (s, "end") ||
         !strcasecmp (s, "to");
}

static void HandleImport (int is_uses)
{
  char *fn = NULL;
  if (ImportFile != NULL)
    {
      char *pf = ImportFile, *pn = fn = xmalloc (strlen (pf) + 1), q = 0;
      while (*pf)
        if (!q)
          {
            while (*pf && *pf != '\'' && *pf != '"')
              pf++;
            if (*pf)
              q = *pf++;
          }
        else if (pf[1] == q && *pf == (q == '\'' ? q : '\\'))
          {
            *pn++ = q;
            pf += 2;
          }
        else if (*pf == q)
          {
            q = 0;
            pf++;
          }
        else
          *pn++ = *pf++;
      *pn = 0;
      if (pn == fn)
        ParseError (is_uses ? "missing file name after `uses ... in'" : "missing file name after `import ... in'");
      free (ImportFile);
    }
  if (/* is_uses || */  /* GPC allows `uses StandardInput' as well; if this ever changes, uncomment this check */
      (strcasecmp (ImportName, "StandardInput") &&
       strcasecmp (ImportName, "StandardOutput") &&
       strcasecmp (ImportName, "StandardError")))
    AddList ((ImplementationPos == 0) ? &InterfaceUses : &ImplementationUses, ImportName, fn);
  if (fn)
    free (fn);
  free (ImportName);
  ImportName = ImportFile = NULL;
}

#define YY_USER_ACTION UserAction ();
static inline void UserAction (void)
{
  if ((*mp_yytext == 'I' || *mp_yytext == 'i') && !strcasecmp (mp_yytext, "implementation"))
    SetImplementation ();
}

%}

WS        [ \t\n]
STRCONST1 '([^']|\n|'')*'|\"([^\"]|\n|\\.)*\"
STRCONST  {STRCONST1}({WS}*"+"?{WS}*{STRCONST1})*

%%

<INITIAL,Uses,Import1,Import,Export1,Export>'  {
                                      int c;
                                      do
                                        c = input ();
                                      while (c != '\'' && c != EOF);
                                    }
<INITIAL,Uses,Import1,Import,Export1,Export>\"  {
                                      int c;
                                      do
                                        {
                                          c = input ();
                                          while (c == '\\')
                                            {
                                              c = input ();
                                              if (c != EOF)
                                                c = input ();
                                            }
                                        }
                                      while (c != '"' && c != EOF);
                                    }
<INITIAL,Uses,AfterUses,UsesIn,Import1,Import,AfterImport,ImportIn,Export1,Export,AfterExport>"{$"|"(*$" PrevState = YY_START; BEGIN (Directive);
[a-z_][a-z0-9_]*                    {
                                      /* The module handling here is not nice ...
                                         we recognize `module Foo interface' and
                                         ignore `module Foo implementation', so
                                         interface and implementation within the
                                         same file are not counted as two
                                         modules. */
                                      if (Parentheses)
                                        ;
                                      else if (ModuleStatus == 1)
                                        {
                                          if (!*ModuleName)
                                            *ModuleName = strdup (mp_yytext);
                                          ModuleStatus = 2;
                                        }
                                      else if (ModuleStatus == 2)
                                        {
                                          if (!strcasecmp (mp_yytext, "interface"))
                                            SetFileKind (fk_Module);
                                          ModuleStatus = 0;
                                        }
                                      else switch (*mp_yytext)
                                      {
                                        case 'p':
                                        case 'P': if (!strcasecmp (mp_yytext, "program"))
                                                    SetFileKind (fk_Program);
                                                  break;
                                        case 'm':
                                        case 'M': if (!strcasecmp (mp_yytext, "module"))
                                                    ModuleStatus = 1;
                                                  break;
                                        case 'u':
                                        case 'U': if (!strcasecmp (mp_yytext, "uses"))
                                                    BEGIN (Uses);
                                                  else if (!strcasecmp (mp_yytext, "unit"))
                                                    SetFileKind (fk_Unit);
                                                  break;
                                        case 'i':
                                        case 'I': if (!strcasecmp (mp_yytext, "import"))
                                                    BEGIN (Import1);
                                                  break;
                                        case 'e':
                                        case 'E': if (!strcasecmp (mp_yytext, "export"))
                                                    BEGIN (Export1);
                                                  break;
                                      }
                                      if (ModuleStatus == 0)
                                        ComponentStarted = 1;
                                    }
end{WS}*\.                          ComponentStarted = 0;
"("                                 Parentheses++;
")"                                 Parentheses--;
;                                   {
                                      if (!Parentheses && ModuleStatus == 2)
                                        {
                                          SetFileKind (fk_Module);
                                          ModuleStatus = 0;
                                        }
                                    }
[0-9][a-z0-9_]*
[^'"{(a-z0-9_();\n]+
.|\n

<Uses>[a-z_][a-z0-9_]*              {
                                      ImportName = strdup (mp_yytext);
                                      BEGIN (AfterUses);
                                    }
<AfterUses>;                        {
                                      if (ImportName == NULL)
                                        /* ParseError ("parse error in `uses' declaration") */ ;
                                      else
                                        HandleImport (1);
                                      BEGIN (INITIAL);
                                    }
<AfterUses>,                        {
                                      if (ImportName == NULL)
                                        /* ParseError ("parse error in `uses' declaration"); */ BEGIN (INITIAL);
                                      else
                                        {
                                          HandleImport (1);
                                          BEGIN (Uses);
                                        }
                                    }
<AfterUses>in/[^a-z0-9_]            BEGIN (UsesIn);
<UsesIn>{STRCONST}                  {
                                      ImportFile = strdup (mp_yytext);
                                      BEGIN (AfterUses);
                                    }

<Import1,Import>[a-z_][a-z0-9_]*    {
                                      if (YY_START != Import1 && starts_decl_part (mp_yytext))
                                        {
                                          yyless (0);
                                          BEGIN (INITIAL);
                                        }
                                      else
                                        {
                                          ImportName = strdup (mp_yytext);
                                          BEGIN (AfterImport);
                                        }
                                    }
<AfterImport>;                      {
                                      if (ImportName == NULL)
                                        /* ParseError ("parse error in `uses' declaration"); */ BEGIN (INITIAL);
                                      else
                                        {
                                          HandleImport (0);
                                          BEGIN (Import);
                                        }
                                    }
<AfterImport>in/[^a-z0-9_]          BEGIN (ImportIn);
<ImportIn>{STRCONST}                {
                                      ImportFile = strdup (mp_yytext);
                                      BEGIN (AfterImport);
                                    }

<Export1,Export>[a-z_][a-z0-9_]*    {
                                      if (YY_START != Export1 && starts_decl_part (mp_yytext))
                                        {
                                          yyless (0);
                                          BEGIN (INITIAL);
                                        }
                                      else
                                        {
                                          AddList (&ExportedInterfaces, mp_yytext, NULL);
                                          BEGIN (AfterExport);
                                        }
                                    }
<AfterExport>;                      BEGIN (Export);

<AfterUses,AfterImport>qualified/[^a-z0-9_]  /* @@ This is too rough */
<AfterUses,AfterImport>only/[^a-z0-9_]
<AfterUses,AfterImport>"("[^)]*")"
<AfterExport>=
<AfterExport>all/[^a-z0-9_]
<AfterExport>"("[^)]*")"

<Uses,AfterUses,UsesIn,Import1,Import,AfterImport,ImportIn,Export1,Export,AfterExport>{WS}+
<Uses,AfterUses,UsesIn>.                /* ParseError ("parse error in `uses' declaration");   */ BEGIN (INITIAL);
<Import1,Import,AfterImport,ImportIn>.  /* ParseError ("parse error in `import' declaration"); */ BEGIN (INITIAL);
<Export1,Export,AfterExport>.           /* ParseError ("parse error in `export' declaration"); */ BEGIN (INITIAL);

<Directive,Directive1,LinkerDirective,ResourceDirective>"}"|"*)"    BEGIN (PrevState);
<Directive>[lL]/{WS}                                                BEGIN (LinkerDirective);
<Directive>[rR]/{WS}                                                BEGIN (ResourceDirective);
<LinkerDirective,ResourceDirective>([^*} \t\n,]|"*"+[^*)} \t\n,])+  AddList (YY_START == LinkerDirective ? &LinkedFiles : &ResourceFiles, mp_yytext, NULL);
<Directive1>,    BEGIN (Directive);
<Directive>.|\n  BEGIN (Directive1);
<Directive1,LinkerDirective,ResourceDirective>.|\n

%%

TFileKind ParseFile (char *_SrcName, char *_Buf, int _BufCount, int *_ImplementationPos,
                     char **_ModuleName, CStringList **_ExportedInterfaces,
                     CStringList **_InterfaceUses, CStringList **_ImplementationUses,
                     CStringList **_LinkedFiles, CStringList **_ResourceFiles)
{
  SrcName = strdup (_SrcName);
  Buf = _Buf;
  BufCount = _BufCount;
  ModuleName = _ModuleName;
  ExportedInterfaces = _ExportedInterfaces;
  InterfaceUses = _InterfaceUses;
  ImplementationUses = _ImplementationUses;
  LinkedFiles = _LinkedFiles;
  ResourceFiles = _ResourceFiles;
  *ModuleName = NULL;
  *ExportedInterfaces = *InterfaceUses = *ImplementationUses = *LinkedFiles = *ResourceFiles = NULL;
  FileKind = fk_None;
  StringChar = 0;
  ImplementationPos = BufPos = 0;
  Parentheses = 0;
  ComponentStarted = 0;
  ModuleStatus = 0;
  Column = Line = 1;
  Comment = Digraph = 0;
  ImportName = ImportFile = NULL;
  PrevState = INITIAL;
  YY_FLUSH_BUFFER;
  mp_yylex ();
  if (BufPos != BufCount) ParseError ("internal error: parser position mismatch");
  if (ImplementationPos == 0) SetImplementation ();
  *_ImplementationPos = ImplementationPos;
  free (SrcName);
  return FileKind;
}
